# __version__.py
# autogenerated by keats 0.2.24
__version__ = "0.5.4"
__name__ = "lobio"
__title__ = "lobio"
__authors__ = ["Justin Vrana <justin.vrana@gmail.com>"]
__repo__ = None
__homepage__ = "https://www.github.com/jvrana/LoBio"
__description__ = "Extra utilities for BioPython"
